# Copyright (c) 2019, Battelle Memorial Institute
# All rights reserved.
#
# See LICENSE.txt and WARRANTY.txt for details.


import shapely.geometry.base
import shapely.geometry.point

from buildingid.code import Code, encode


class DictDatum:
    def __init__(
        self,
        geom: shapely.geometry.base.BaseGeometry,
        bounds: tuple[float, float, float, float] | None = None,
        centroid: shapely.geometry.point.Point | None = None,
    ) -> None:
        super().__init__()

        self.geom = geom

        self._bounds = bounds
        self._centroid = centroid

    @property
    def bounds(self) -> tuple[float, float, float, float]:
        if self._bounds is None:
            return self.geom.bounds
        else:
            return self._bounds

    @property
    def centroid(self) -> shapely.geometry.point.Point:
        if self._centroid is None:
            return self.geom.centroid
        else:
            return self._centroid

    def encode(self, **kwargs) -> Code:
        bounds = self.bounds
        centroid = self.centroid

        return encode(bounds[1], bounds[0], bounds[3], bounds[2], centroid.y, centroid.x, **kwargs)



# Copyright (c) 2019, Battelle Memorial Institute
# All rights reserved.
#
# See LICENSE.txt and WARRANTY.txt for details.

import typing

from openlocationcode import openlocationcode

from buildingid.code_pattern import (
    FORMAT_STRING_,
    RE_GROUP_EAST_,
    RE_GROUP_NORTH_,
    RE_GROUP_OPENLOCATIONCODE_,
    RE_GROUP_SOUTH_,
    RE_GROUP_WEST_,
    RE_PATTERN_,
)
from buildingid.validators import (
    isValidCodeArea,
    isValidCodeLength,
    isValidLatitude,
    isValidLatitudeCenter,
    isValidLongitude,
    isValidLongitudeCenter,
)

Code = typing.NewType("Code", str)


class CodeArea(openlocationcode.CodeArea):
    def __init__(self, centroid: openlocationcode.CodeArea, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.centroid = centroid

    @property
    def area(self) -> float:
        height = float(self.latitudeHi - self.latitudeLo)
        width = float(self.longitudeHi - self.longitudeLo)

        return height * width

    def encode(self) -> Code:
        return encode(
            self.latitudeLo,
            self.longitudeLo,
            self.latitudeHi,
            self.longitudeHi,
            self.centroid.latitudeCenter,
            self.centroid.longitudeCenter,
            codeLength=self.codeLength,
        )

    def intersection(self, other: "CodeArea") -> tuple[float, float, float, float] | None:
        if other is None:
            return None

        latitudeLo = float(max(self.latitudeLo, other.latitudeLo))
        latitudeHi = float(min(self.latitudeHi, other.latitudeHi))
        longitudeLo = float(max(self.longitudeLo, other.longitudeLo))
        longitudeHi = float(min(self.longitudeHi, other.longitudeHi))

        if (latitudeLo > latitudeHi) or (longitudeLo > longitudeHi):
            return None

        return (
            longitudeLo,
            latitudeLo,
            longitudeHi,
            latitudeHi,
        )

    def jaccard(self, other: "CodeArea") -> float | None:
        bbox = self.intersection(other)

        if bbox is None:
            return None

        area = (bbox[3] - bbox[1]) * (bbox[2] - bbox[0])

        return area / (self.area + other.area - area)

    def resize(self) -> "CodeArea":
        halfHeight = float(self.centroid.latitudeHi - self.centroid.latitudeLo) / 2
        halfWidth = float(self.centroid.longitudeHi - self.centroid.longitudeLo) / 2

        return CodeArea(
            self.centroid,
            self.latitudeLo + halfHeight,
            self.longitudeLo + halfWidth,
            self.latitudeHi - halfHeight,
            self.longitudeHi - halfWidth,
            codeLength=self.codeLength,
        )


def decode(code: Code) -> CodeArea:
    match = isValid_(code)

    if match is None:
        raise ValueError("buildingid.code.decode - Invalid code")

    codeAreaCenter = openlocationcode.decode(match.group(RE_GROUP_OPENLOCATIONCODE_))
    assert isValidCodeArea(codeAreaCenter), "buildingid.code.decode - Invalid code area"

    codeAreaCenterHeight = codeAreaCenter.latitudeHi - codeAreaCenter.latitudeLo
    assert codeAreaCenterHeight >= 0, "buildingid.code.decode - Invalid code - Negative height"

    codeAreaCenterWidth = codeAreaCenter.longitudeHi - codeAreaCenter.longitudeLo
    assert codeAreaCenterWidth >= 0, "buildingid.code.decode - Invalid code - Negative width"

    latitudeLo = codeAreaCenter.latitudeLo - (int(match.group(RE_GROUP_SOUTH_)) * codeAreaCenterHeight)
    latitudeHi = codeAreaCenter.latitudeHi + (int(match.group(RE_GROUP_NORTH_)) * codeAreaCenterHeight)
    assert isValidLatitude(latitudeLo, latitudeHi), "buildingid.code.decode - Invalid code - Invalid latitude coordinates"

    longitudeLo = codeAreaCenter.longitudeLo - (int(match.group(RE_GROUP_WEST_)) * codeAreaCenterWidth)
    longitudeHi = codeAreaCenter.longitudeHi + (int(match.group(RE_GROUP_EAST_)) * codeAreaCenterWidth)
    assert isValidLongitude(longitudeLo, longitudeHi), "buildingid.code.decode - Invalid code - Invalid longitude coordinates"

    return CodeArea(codeAreaCenter, latitudeLo, longitudeLo, latitudeHi, longitudeHi, codeAreaCenter.codeLength)


def encode(
    latitudeLo: float,
    longitudeLo: float,
    latitudeHi: float,
    longitudeHi: float,
    latitudeCenter: float,
    longitudeCenter: float,
    codeLength: int = openlocationcode.PAIR_CODE_LENGTH_,
) -> Code:
    assert isValidCodeLength(codeLength), "buildingid.code.encode - Invalid code length"
    assert isValidLatitudeCenter(latitudeLo, latitudeHi, latitudeCenter), "buildingid.code.encode - Invalid latitude coordinates"
    assert isValidLongitudeCenter(longitudeLo, longitudeHi, longitudeCenter), "buildingid.code.encode - Invalid longitude coordinates"

    olcNortheast = openlocationcode.encode(latitudeHi, longitudeHi, codeLength=codeLength)
    codeAreaNortheast = openlocationcode.decode(olcNortheast)
    assert isValidCodeArea(codeAreaNortheast), "buildingid.code.encode - Invalid code area (northeast)"

    olcSouthwest = openlocationcode.encode(latitudeLo, longitudeLo, codeLength=codeLength)
    codeAreaSouthwest = openlocationcode.decode(olcSouthwest)
    assert isValidCodeArea(codeAreaSouthwest), "buildingid.code.encode - Invalid code area (southwest)"

    olcCenter = openlocationcode.encode(latitudeCenter, longitudeCenter, codeLength=codeLength)
    codeAreaCenter = openlocationcode.decode(olcCenter)
    assert isValidCodeArea(codeAreaCenter), "buildingid.code.encode - Invalid code area (center)"

    codeAreaCenterHeight = codeAreaCenter.latitudeHi - codeAreaCenter.latitudeLo
    assert codeAreaCenterHeight >= 0, "buildingid.code.encode - Invalid code - Negative height"

    codeAreaCenterWidth = codeAreaCenter.longitudeHi - codeAreaCenter.longitudeLo
    assert codeAreaCenterWidth >= 0, "buildingid.code.encode - Invalid code - Negative width"

    olcCountNorth = (codeAreaNortheast.latitudeHi - codeAreaCenter.latitudeHi) / codeAreaCenterHeight
    assert olcCountNorth >= 0, "buildingid.code.encode - Negative extent (north)"

    olcCountSouth = (codeAreaCenter.latitudeLo - codeAreaSouthwest.latitudeLo) / codeAreaCenterHeight
    assert olcCountSouth >= 0, "buildingid.code.encode - Negative extent (south)"

    olcCountEast = (codeAreaNortheast.longitudeHi - codeAreaCenter.longitudeHi) / codeAreaCenterWidth
    assert olcCountEast >= 0, "buildingid.code.encode - Negative extent (east)"

    olcCountWest = (codeAreaCenter.longitudeLo - codeAreaSouthwest.longitudeLo) / codeAreaCenterWidth
    assert olcCountWest >= 0, "buildingid.code.encode - Negative extent (west)"

    return Code(FORMAT_STRING_ % (olcCenter, olcCountNorth, olcCountEast, olcCountSouth, olcCountWest))


def isValid(code: Code) -> bool:
    return isValid_(code) is not None


def isValid_(code: Code) -> typing.Match[str] | None:
    if code is None:
        return None

    match = RE_PATTERN_.match(str(code))

    if (match is None) or not openlocationcode.isValid(match.group(RE_GROUP_OPENLOCATIONCODE_)):
        return None

    return match



# Copyright (c) 2019, Battelle Memorial Institute
# All rights reserved.
#
# See LICENSE.txt and WARRANTY.txt for details.

import typing

from buildingid.command_line.dict_datum import DictDatum
from buildingid.command_line.dict_pipe import DictEncoder


class BaseGeometryDictEncoder(DictEncoder[DictDatum]):
    def __init__(self, fieldname_code: str, code_length: int) -> None:
        super().__init__()

        self.fieldname_code = fieldname_code

        self.code_length = code_length

    def encode(self, datum: DictDatum) -> dict[str, typing.Any]:
        code = datum.encode(codeLength=self.code_length)

        row = {}

        row[self.fieldname_code] = code

        return row

    @property
    def fieldnames(self) -> list[str]:
        return [
            self.fieldname_code,
        ]


class ErrorDictEncoder(DictEncoder[BaseException]):
    def __init__(self, fieldname_code: str) -> None:
        super().__init__()

        self.fieldname_code = fieldname_code

    def encode(self, exception: BaseException) -> dict[str, typing.Any]:
        row = {}

        row[f"{self.fieldname_code}_Error_Name"] = type(exception).__name__
        row[f"{self.fieldname_code}_Error_Message"] = str(exception)

        return row

    @property
    def fieldnames(self) -> list[str]:
        return [
            f"{self.fieldname_code}_Error_Name",
            f"{self.fieldname_code}_Error_Message",
        ]



# Copyright (c) 2019, Battelle Memorial Institute
# All rights reserved.
#
# See LICENSE.txt and WARRANTY.txt for details.

import abc
import csv
import typing

from buildingid.command_line.exceptions import FieldNotFoundError, FieldNotUniqueError

T = typing.TypeVar("T")


class DictDecoder(abc.ABC, typing.Generic[T]):
    def __init__(self) -> None:
        super().__init__()

    @abc.abstractmethod
    def decode(self, row: dict[str, str]) -> T:
        raise NotImplementedError()  # pragma: no cover

    @property
    @abc.abstractmethod
    def fieldnames(self) -> list[str]:
        raise NotImplementedError()  # pragma: no cover


class DictEncoder(abc.ABC, typing.Generic[T]):
    def __init__(self) -> None:
        super().__init__()

    @abc.abstractmethod
    def encode(self, inst: T) -> dict[str, typing.Any]:
        raise NotImplementedError()  # pragma: no cover

    @property
    @abc.abstractmethod
    def fieldnames(self) -> list[str]:
        raise NotImplementedError()  # pragma: no cover


class DictPipe:
    def __init__(self, decoder_in: DictDecoder[T], encoder_out: DictEncoder[T], encoder_err: DictEncoder[BaseException]) -> None:
        super().__init__()

        self.decoder_in = decoder_in
        self.encoder_out = encoder_out
        self.encoder_err = encoder_err

    def run(
        self,
        io_in: typing.TextIO,
        io_out: typing.TextIO,
        io_err: typing.TextIO,
        args_in: list = [],
        kwargs_in: dict = {},
        args_out: list = [],
        kwargs_out: dict = {},
    ) -> None:
        csv_in = csv.DictReader(io_in, *args_in, **kwargs_in)

        fieldnames_in = csv_in.fieldnames
        fieldnames_in = [] if (fieldnames_in is None) else list(fieldnames_in)

        for fieldname in self.decoder_in.fieldnames:
            if fieldname not in fieldnames_in:
                raise FieldNotFoundError(fieldname)

        fieldnames_out = fieldnames_in.copy()

        for fieldname in self.encoder_out.fieldnames:
            if fieldname in fieldnames_out:
                raise FieldNotUniqueError(fieldname)

            fieldnames_out.append(fieldname)

        fieldnames_err = fieldnames_in.copy()

        for fieldname in self.encoder_err.fieldnames:
            if fieldname in fieldnames_err:
                raise FieldNotUniqueError(fieldname)

            fieldnames_err.append(fieldname)

        csv_out = csv.DictWriter(io_out, fieldnames_out, *args_out, **kwargs_out)
        csv_err = csv.DictWriter(io_err, fieldnames_err, *args_out, **kwargs_out)

        csv_out_writeheader_called = False
        csv_err_writeheader_called = False

        for in_row in csv_in:
            try:
                inst = self.decoder_in.decode(in_row)

                out_row = in_row.copy()
                out_row.update(self.encoder_out.encode(inst))
            except BaseException as exception:
                err_row = in_row.copy()
                err_row.update(self.encoder_err.encode(exception))

                if not csv_err_writeheader_called:
                    csv_err_writeheader_called = True

                    csv_err.writeheader()

                csv_err.writerow(err_row)
            else:
                if not csv_out_writeheader_called:
                    csv_out_writeheader_called = True

                    csv_out.writeheader()

                csv_out.writerow(out_row)



# Copyright (c) 2019, Battelle Memorial Institute
# All rights reserved.
#
# See LICENSE.txt and WARRANTY.txt for details.

import importlib.metadata
import logging
import typing

import click
import click_log
import pandas as pd
import pyqtree
from tldm import pandas, tldm

from buildingid.code import Code, CodeArea, decode
from buildingid.command_line.dict_decoders import LatLngDictDecoder, WKBDictDecoder, WKTDictDecoder
from buildingid.command_line.dict_encoders import BaseGeometryDictEncoder, ErrorDictEncoder
from buildingid.command_line.dict_pipe import DictDecoder, DictPipe
from buildingid.command_line.exceptions import CustomError, FieldNotFoundError, FieldNotUniqueError
from buildingid.command_line.set_csv_field_size_limit import set_csv_field_size_limit
from buildingid.validators import isValidCodeLength

pandas()

VERSION = importlib.metadata.version("buildingid")
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

click_log.basic_config(logger)

set_csv_field_size_limit()


def click_callback_code_length_(ctx: None, opt: click.core.Option, codeLength: int) -> int:
    """Callback for "--code-length" option (the number of digits in the OLC segment of the UBID string).

    See `buildingid.validators.isValidCodeLength` for details.
    """

    if isValidCodeLength(codeLength):
        return codeLength
    else:
        raise click.BadParameter(f"Invalid Open Location Code length: {codeLength!s}")


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.pass_context
@click.version_option(VERSION)
def cli(ctx: None) -> None:
    """Unique Building Identifier (UBID) Software (Battelle IPID 31307-E)

    Unique Building Identifier (UBID) pinpoints the exact physical location and extent of space that a building occupies.
    These details make it possible to gather and study accurate, reliable information about a building and its spatial characteristics.
    """


@cli.command("append2csv", short_help='append "UBID" field to rows of CSV file')
@click.argument("dict-decoder-id", nargs=1, type=click.Choice(["latlng", "wkb", "wkt"], case_sensitive=True))
@click.option(
    "--code-length",
    type=click.IntRange(0, None),
    default=11,
    show_default=True,
    callback=click_callback_code_length_,
    help="the number of digits in the OLC segment of the UBID string",
)
@click.option(
    "--fieldname-south-latitude",
    type=click.STRING,
    default=None,
    show_default=True,
    help="the name of the south latitude field in the input file",
)
@click.option(
    "--fieldname-west-longitude",
    type=click.STRING,
    default=None,
    show_default=True,
    help="the name of the west longitude field in the input file",
)
@click.option(
    "--fieldname-north-latitude",
    type=click.STRING,
    default=None,
    show_default=True,
    help="the name of the north latitude field in the input file",
)
@click.option(
    "--fieldname-east-longitude",
    type=click.STRING,
    default=None,
    show_default=True,
    help="the name of the east longitude field in the input file",
)
@click.option(
    "--fieldname-center-latitude",
    type=click.STRING,
    default="Latitude",
    show_default=True,
    help="the name of the center latitude field in the input file",
)
@click.option(
    "--fieldname-center-longitude",
    type=click.STRING,
    default="Longitude",
    show_default=True,
    help="the name of the center longitude field in the input file",
)
@click.option(
    "--fieldname-wkbstr",
    type=click.STRING,
    default="WKB",
    show_default=True,
    help="the name of the hex-encoded well-known binary (WKB) string field in the input file",
)
@click.option(
    "--fieldname-wktstr",
    type=click.STRING,
    default="WKT",
    show_default=True,
    help="the name of the well-known text (WKT) string field in the input file",
)
@click.option(
    "--fieldname-code",
    type=click.STRING,
    default="UBID",
    show_default=True,
    help="the name of the UBID field in the output file",
)
@click.option("--reader-delimiter", type=click.STRING, default=",", show_default=True, help="the delimiter to use for the input file")
@click.option(
    "--reader-quotechar",
    type=click.STRING,
    default='"',
    show_default=True,
    help="the character used to denote the start and end of a quoted field in the input file",
)
@click.option("--writer-delimiter", type=click.STRING, default=",", show_default=True, help="the delimiter to use for the output file")
@click.option(
    "--writer-quotechar",
    type=click.STRING,
    default='"',
    show_default=True,
    help="the character used to denote the start and end of a quoted field in the output file",
)
@click.pass_context
def run_append_to_csv(
    ctx: None,
    dict_decoder_id: str,
    code_length: int,
    fieldname_south_latitude: str,
    fieldname_west_longitude: str,
    fieldname_north_latitude: str,
    fieldname_east_longitude: str,
    fieldname_center_latitude: str,
    fieldname_center_longitude: str,
    fieldname_wkbstr: str,
    fieldname_wktstr: str,
    fieldname_code: str,
    reader_delimiter: str,
    reader_quotechar: str,
    writer_delimiter: str,
    writer_quotechar: str,
) -> None:
    """The \033[1mappend2csv\033[0m command assigns a Unique Building Identifier (UBID) to each row in the input file.

    The input, output and error files are represented in comma-separated values (CSV) format.  The input file is read from the standard input stream.  The output file is written to the standard output stream.  The error file is written to the standard error stream.

    For each row in the input file, the geometry for the row is assigned a UBID.  If a UBID is successfully assigned, then the row is written to the output file with an added "UBID" field (the \033[1m--fieldname-code\033[0m option).  If a UBID is not assigned, for example, if an exception is raised, then the row is written to the error file.

    The following modes are available, where each mode corresponds to a geometry type:

    \033[1mlatlng\033[0m\tPoints; represented as latitude and longitude coordinates, read from the "Latitude" and "Longitude" fields (the \033[1m--fieldname-center-latitude\033[0m and \033[1m--fieldname-center-longitude\033[0m options).

    \033[1mwkb\033[0m\t\tShapes; represented as strings in hex-encoded well-known binary (WKB) format read from the "WKB" field (the \033[1m--fieldname-wkbstr\033[0m option).

    \033[1mwkt\033[0m\t\tShapes; represented as strings in well-known text (WKT) format, read from the "WKT" field (the \033[1m--fieldname-wktstr\033[0m option).

    The number of digits in the Open Location Code (OLC) segment of the UBID string is specified by the \033[1m--code-length\033[0m option.

    The \033[1mappend2csv\033[0m command exits 0 on success, and >0 if an error occurs.
    """

    # Construct `DictDecoder[DictDatum]` for standard input stream.
    decoder_in: DictDecoder
    if dict_decoder_id == "latlng":
        decoder_in = LatLngDictDecoder(
            fieldname_center_latitude,
            fieldname_center_longitude,
            fieldname_north_latitude=fieldname_north_latitude,
            fieldname_south_latitude=fieldname_south_latitude,
            fieldname_east_longitude=fieldname_east_longitude,
            fieldname_west_longitude=fieldname_west_longitude,
        )
    elif dict_decoder_id == "wkb":
        decoder_in = WKBDictDecoder(fieldname_wkbstr)
    elif dict_decoder_id == "wkt":
        decoder_in = WKTDictDecoder(fieldname_wktstr)
    else:
        pass

    # Construct `DictEncoder[DictDatum]` for standard output stream.
    encoder_out = BaseGeometryDictEncoder(fieldname_code, code_length)

    # Construct `DictEncoder[BaseException]` for standard error stream.
    encoder_err = ErrorDictEncoder(fieldname_code)

    # Construct `DictPipe` for standard input, output and error streams.
    dict_pipe = DictPipe(decoder_in, encoder_out, encoder_err)

    # Standard input, output and error streams.
    io_in = click.get_text_stream("stdin")
    io_out = click.get_text_stream("stdout")
    io_err = click.get_text_stream("stderr")

    # Configuration for `csv.DictReader`.
    args_in: list = []
    kwargs_in = {
        "delimiter": reader_delimiter,
        "quotechar": reader_quotechar,
    }

    # Configuration for `csv.DictWriter`.
    args_out: list = []
    kwargs_out = {
        "delimiter": writer_delimiter,
        "quotechar": writer_quotechar,
        # 'quoting': csv.QUOTE_NONNUMERIC,
    }

    try:
        dict_pipe.run(io_in, io_out, io_err, args_in=args_in, kwargs_in=kwargs_in, args_out=args_out, kwargs_out=kwargs_out)
    except CustomError as exception:
        raise click.ClickException(exception)

    # Done!


@cli.command("crossref", short_help='cross-reference "UBID" fields in rows of two CSV files')
@click.argument("left", type=click.File("r"))
@click.argument("right", type=click.File("r"))
@click.argument("dst", type=click.File("w"))
@click.option(
    "--fieldname-jaccard",
    type=click.STRING,
    default="IoU",
    show_default=True,
    help='the name of the Jaccard similarity coefficient (viz., "intersection over union" or "IoU") in the output file',
)
@click.option(
    "--include-jaccard-field",
    is_flag=True,
    default=False,
    show_default=True,
    help="include the Jaccard similarity coefficient in the output file",
)
@click.option("--include-index-fields", is_flag=True, default=False, show_default=True, help="include the index field in the output file")
@click.option(
    "--include-left-field",
    type=click.STRING,
    multiple=True,
    help="include the named field of the left input file in the output file",
)
@click.option(
    "--include-right-field",
    type=click.STRING,
    multiple=True,
    help="include the named field of the right input file in the output file",
)
@click.option(
    "--jaccard-min",
    type=click.FloatRange(min=0.0, max=1.0),
    default=0.0,
    show_default=True,
    help="the minimum value of the Jaccard similarity coefficient",
)
@click.option(
    "--jaccard-max",
    type=click.FloatRange(min=0.0, max=1.0),
    default=1.0,
    show_default=True,
    help="the maximum value of the Jaccard similarity coefficient",
)
@click.option(
    "--sort-by-jaccard",
    is_flag=True,
    default=False,
    show_default=True,
    help="sort the rows of the output file by the Jaccard similarity coefficient",
)
@click.option(
    "--sort-order",
    type=click.Choice(["ASC", "DESC"], case_sensitive=True),
    default="ASC",
    show_default=True,
    help="the sort order for the rows of the output file",
)
@click.option(
    "--left-group-by-jaccard",
    is_flag=True,
    default=False,
    show_default=True,
    help="group the rows of the left input file by their UBID strings",
)
@click.option(
    "--left-group-order",
    type=click.Choice(["ASC", "DESC"], case_sensitive=True),
    default="ASC",
    show_default=True,
    help="the sort order for the groups of rows of the left input file",
)
@click.option(
    "--right-group-by-jaccard",
    is_flag=True,
    default=False,
    show_default=True,
    help="group the rows of the right input file by their UBID string",
)
@click.option(
    "--right-group-order",
    type=click.Choice(["ASC", "DESC"], case_sensitive=True),
    default="ASC",
    show_default=True,
    help="the sort order for the groups of rows of the right input file",
)
@click.option(
    "--left-fieldname-code",
    type=click.STRING,
    default="UBID",
    show_default=True,
    help="the name of the UBID field in the left input file",
)
@click.option(
    "--left-fieldname-index",
    type=click.STRING,
    default="index",
    show_default=True,
    help="the name of the index field in the left input file",
)
@click.option(
    "--left-fieldname-openlocationcode",
    type=click.STRING,
    default="__openlocationcode__",
    show_default=True,
    help="the name of the temporary field for decoded UBID strings in the left input file",
)
@click.option("--left-suffix", type=click.STRING, default="_x", show_default=True, help="the suffix for field names in the left input file")
@click.option(
    "--left-reader-delimiter",
    type=click.STRING,
    default=",",
    show_default=True,
    help="the delimiter to use for the left input file",
)
@click.option(
    "--left-reader-quotechar",
    type=click.STRING,
    default='"',
    show_default=True,
    help="the character used to denote the start and end of a quoted field in the left input file",
)
@click.option(
    "--right-fieldname-code",
    type=click.STRING,
    default="UBID",
    show_default=True,
    help="the name of the UBID field in the right input file",
)
@click.option(
    "--right-fieldname-index",
    type=click.STRING,
    default="index",
    show_default=True,
    help="the name of the index field in the right input file",
)
@click.option(
    "--right-fieldname-openlocationcode",
    type=click.STRING,
    default="__openlocationcode__",
    show_default=True,
    help="the name of the temporary field for decoded UBID strings in the right input file",
)
@click.option(
    "--right-suffix",
    type=click.STRING,
    default="_y",
    show_default=True,
    help="the suffix for field names in the right input file",
)
@click.option(
    "--right-reader-delimiter",
    type=click.STRING,
    default=",",
    show_default=True,
    help="the delimiter to use for the right input file",
)
@click.option(
    "--right-reader-quotechar",
    type=click.STRING,
    default='"',
    show_default=True,
    help="the character used to denote the start and end of a quoted field in the right input file",
)
@click.option("--writer-delimiter", type=click.STRING, default=",", show_default=True, help="the delimiter to use for the output file")
@click.option(
    "--writer-quotechar",
    type=click.STRING,
    default='"',
    show_default=True,
    help="the character used to denote the start and end of a quoted field in the output file",
)
@click.pass_context
def run_crossref(
    ctx: None,
    left: typing.TextIO,
    right: typing.TextIO,
    dst: typing.TextIO,
    fieldname_jaccard: str,
    include_jaccard_field: bool,
    include_index_fields: bool,
    include_left_field: list[str],
    include_right_field: list[str],
    jaccard_min: float,
    jaccard_max: float,
    sort_by_jaccard: bool,
    sort_order: str,
    left_group_by_jaccard: bool,
    left_group_order: str,
    right_group_by_jaccard: bool,
    right_group_order: str,
    left_fieldname_code: str,
    left_fieldname_index: str,
    left_fieldname_openlocationcode: str,
    left_suffix: str,
    left_reader_delimiter: str,
    left_reader_quotechar: str,
    right_fieldname_code: str,
    right_fieldname_index: str,
    right_fieldname_openlocationcode: str,
    right_suffix: str,
    right_reader_delimiter: str,
    right_reader_quotechar: str,
    writer_delimiter: str,
    writer_quotechar: str,
) -> None:
    """The \033[1mcrossref\033[0m command cross-references the Unique Building Identifiers (UBIDs) in the rows of the left and right input files.

    The left input, right input and output files are represented in comma-separated values (CSV) format.  The left and right input files are read from either the standard input stream (using "-") or the specified path.  The output file is written to either the standard output stream (using "-") or the specified path.

    The larger of the two input files is used to construct a quadtree-based spatial index.  The smaller of the two input files is traversed, row at a time, to identify intersecting UBID bounding boxes.  For each intersection, the Jaccard similarity coefficient (viz., "intersection over union" or "IoU") is calculated, and the row is written to the output file.

    The \033[1mcrossref\033[0m command exits 0 on success, and >0 if an error occurs.
    """

    # Configuration for `pd.read_csv` for left input file.
    kwargs_for_read_csv_left: dict[str, typing.Any] = {
        "dtype": {
            left_fieldname_code: str,
        },
        "quotechar": left_reader_quotechar,
        "sep": left_reader_delimiter,
        "usecols": [left_fieldname_code, *list(include_left_field)],
    }

    # Configuration for `pd.read_csv` for right input file.
    kwargs_for_read_csv_right: dict[str, typing.Any] = {
        "dtype": {
            right_fieldname_code: str,
        },
        "quotechar": right_reader_quotechar,
        "sep": right_reader_delimiter,
        "usecols": [right_fieldname_code, *list(include_right_field)],
    }

    # Configuration for `pd.to_csv` for output file.
    kwargs_for_to_csv_dst: dict[str, typing.Any] = {
        "quotechar": writer_quotechar,
        # 'quoting': csv.QUOTE_NONNUMERIC,
        "sep": writer_delimiter,
    }

    # Names for "index" fields for left and right input files.
    left_fieldname_index_with_suffix: str = f"{left_fieldname_index}{left_suffix}"
    right_fieldname_index_with_suffix: str = f"{right_fieldname_index}{right_suffix}"

    # Names for "__openlocationcode__" fields for left and right input files.
    left_fieldname_openlocationcode_with_suffix: str = f"{left_fieldname_openlocationcode}{left_suffix}"
    right_fieldname_openlocationcode_with_suffix: str = f"{right_fieldname_openlocationcode}{right_suffix}"

    def callback_decode_(code: Code) -> CodeArea | None:
        """Return the decoded 'CodeArea' for the given 'Code', or 'None' if an error occurs."""

        if code is None:
            return None

        try:
            codeArea = decode(code)
        except (
            AssertionError,
            ValueError,
        ):
            return None

        return codeArea

    def sort_order_to_ascending_(value: str) -> bool:
        """Return the "ascending" argument for the `pd.DataFrame.sort_values` method."""

        return value != "DESC"

    try:
        # Construct 'pd.DataFrame' for left input file.
        #
        # Ensure that "UBID" field is present and that "index" and "__openlocationcode__" fields are not present.
        #
        # Finally, construct 'pd.Series' for 'CodeArea' by decoding "UBID" field.
        logger.info(f"[crossref] Reading left input file: {left.name!r}")
        left_data_frame: pd.DataFrame = pd.read_csv(filepath_or_buffer=left, **kwargs_for_read_csv_left)
        if left_fieldname_code not in left_data_frame:
            raise FieldNotFoundError(left_fieldname_code)
        elif left_fieldname_index_with_suffix in left_data_frame:
            raise FieldNotUniqueError(left_fieldname_index_with_suffix)
        elif left_fieldname_openlocationcode_with_suffix in left_data_frame:
            raise FieldNotUniqueError(left_fieldname_openlocationcode_with_suffix)
        left_series_codeArea: pd.Series = left_data_frame[left_fieldname_code].progress_apply(callback_decode_)
        logger.info(
            f"[crossref] Decoded \033[1m{left_series_codeArea.count()}/{len(left_data_frame)} ({round((left_series_codeArea.count() / len(left_data_frame)) * 100, 2)}%)\033[0m rows of left input file",
        )

        # Construct 'pd.DataFrame' for right input file.
        #
        # Ensure that "UBID" field is present and that "index" and "__openlocationcode__" fields are not present.
        #
        # Finally, construct 'pd.Series' for 'CodeArea' by decoding "UBID" field.
        logger.info(f"[crossref] Reading right input file: {right.name!r}")
        right_data_frame: pd.DataFrame = pd.read_csv(filepath_or_buffer=right, **kwargs_for_read_csv_right)
        if right_fieldname_code not in right_data_frame:
            raise FieldNotFoundError(right_fieldname_code)
        elif right_fieldname_index_with_suffix in right_data_frame:
            raise FieldNotUniqueError(right_fieldname_index_with_suffix)
        elif right_fieldname_openlocationcode_with_suffix in right_data_frame:
            raise FieldNotUniqueError(right_fieldname_openlocationcode_with_suffix)
        right_series_codeArea: pd.Series = right_data_frame[right_fieldname_code].progress_apply(callback_decode_)
        logger.info(
            f"[crossref] Decoded \033[1m{right_series_codeArea.count()}/{len(right_data_frame)} ({round((right_series_codeArea.count() / len(right_data_frame)) * 100, 2)}%)\033[0m rows of right input file",
        )

        # Construct quadtree-based spatial index.
        dst_data: list[tuple[int, int, CodeArea, CodeArea]]
        if len(left_series_codeArea) >= len(right_series_codeArea):
            # If left input file has more rows than right input file, then construct
            # spatial index using left input file and cross-reference with rows of
            # right input file.

            logger.info("[crossref] Constructing quadtree for left input file")
            left_spindex: pyqtree.Index = pyqtree.Index(
                bbox=[
                    left_series_codeArea.progress_apply(lambda codeArea: codeArea.longitudeLo).min(),
                    left_series_codeArea.progress_apply(lambda codeArea: codeArea.latitudeLo).min(),
                    left_series_codeArea.progress_apply(lambda codeArea: codeArea.longitudeHi).max(),
                    left_series_codeArea.progress_apply(lambda codeArea: codeArea.latitudeHi).max(),
                ],
            )
            for left_index, left_codeArea in tldm(left_series_codeArea.items(), total=len(left_series_codeArea)):
                if left_codeArea is not None:
                    left_spindex.insert(
                        item=(left_index, left_codeArea),
                        bbox=[left_codeArea.longitudeLo, left_codeArea.latitudeLo, left_codeArea.longitudeHi, left_codeArea.latitudeHi],
                    )

            logger.info("[crossref] Cross-referencing rows of right input file against quadtree for left input file")
            dst_data = [
                (left_index, right_index, left_codeArea, right_codeArea)
                for right_index, right_codeArea in tldm(right_series_codeArea.items(), total=len(right_series_codeArea))
                if right_codeArea is not None
                for (left_index, left_codeArea) in left_spindex.intersect(
                    [right_codeArea.longitudeLo, right_codeArea.latitudeLo, right_codeArea.longitudeHi, right_codeArea.latitudeHi],
                )
                if left_codeArea is not None
            ]
        else:
            # If right input file has more rows than left input file, then construct
            # spatial index using right input file and cross-reference with rows of
            # left input file.

            logger.info("[crossref] Constructing quadtree for right input file")
            right_spindex: pyqtree.Index = pyqtree.Index(
                bbox=[
                    right_series_codeArea.progress_apply(lambda codeArea: codeArea.longitudeLo).min(),
                    right_series_codeArea.progress_apply(lambda codeArea: codeArea.latitudeLo).min(),
                    right_series_codeArea.progress_apply(lambda codeArea: codeArea.longitudeHi).max(),
                    right_series_codeArea.progress_apply(lambda codeArea: codeArea.latitudeHi).max(),
                ],
            )
            for right_index, right_codeArea in tldm(right_series_codeArea.items(), total=len(right_series_codeArea)):
                if right_codeArea is not None:
                    right_spindex.insert(
                        item=(right_index, right_codeArea),
                        bbox=[right_codeArea.longitudeLo, right_codeArea.latitudeLo, right_codeArea.longitudeHi, right_codeArea.latitudeHi],
                    )

            logger.info("[crossref] Cross-referencing rows of left input file against quadtree for right input file")
            dst_data = [
                (left_index, right_index, left_codeArea, right_codeArea)
                for left_index, left_codeArea in tldm(left_series_codeArea.items(), total=len(left_series_codeArea))
                if left_codeArea is not None
                for (right_index, right_codeArea) in right_spindex.intersect(
                    [left_codeArea.longitudeLo, left_codeArea.latitudeLo, left_codeArea.longitudeHi, left_codeArea.latitudeHi],
                )
                if right_codeArea is not None
            ]

        # Construct 'pd.DataFrame' for cross-reference results (viz., "intersections").
        dst_data_frame: pd.DataFrame = pd.DataFrame(
            data=dst_data,
            columns=[
                left_fieldname_index_with_suffix,
                right_fieldname_index_with_suffix,
                left_fieldname_openlocationcode_with_suffix,
                right_fieldname_openlocationcode_with_suffix,
            ],
        )

        # If there are no cross-reference results, then exit.
        len_dst_data_frame0: int = len(dst_data_frame)
        logger.info(f"[crossref] Found \033[1m{len_dst_data_frame0}\033[0m intersection{'' if len_dst_data_frame0 == 1 else 's'}")
        if len_dst_data_frame0 == 0:
            return

        # Calculate Jaccard similarity coefficient (viz., "intersection over union", "IoU", etc.) for each cross-reference result.
        logger.info(f"[crossref] Calculating field: {fieldname_jaccard!r}")
        if fieldname_jaccard in dst_data_frame.columns:
            raise FieldNotUniqueError(fieldname_jaccard)
        dst_data_frame[fieldname_jaccard] = dst_data_frame.progress_apply(
            lambda row: row[left_fieldname_openlocationcode_with_suffix].jaccard(row[right_fieldname_openlocationcode_with_suffix]),  # type: ignore[operator]
            axis=1,
        )

        # Select cross-reference results within the specified open interval.
        logger.info(f"[crossref] Filtering intersections: {jaccard_min} <= {fieldname_jaccard!r} <= {jaccard_max}")
        dst_data_frame = dst_data_frame[
            dst_data_frame[fieldname_jaccard].notna()
            & (jaccard_min <= dst_data_frame[fieldname_jaccard])
            & (dst_data_frame[fieldname_jaccard] <= jaccard_max)
        ]

        # If there are no cross-reference results, then exit.
        len_dst_data_frame1: int = len(dst_data_frame)
        logger.info(
            f"[crossref] Found \033[1m{len_dst_data_frame1}/{len_dst_data_frame0} ({round((len_dst_data_frame1 / len_dst_data_frame0) * 100, 2)}%)\033[0m intersection{'' if len_dst_data_frame0 == 1 else 's'}: {fieldname_jaccard!r}",
        )
        if len_dst_data_frame1 == 0:
            return

        # Merge left and right input files with cross-reference results using an inner join.
        logger.info("[crossref] Merging intersections with left input file")
        dst_data_frame = dst_data_frame.merge(
            left_data_frame.reset_index().rename(
                index=str,
                columns={
                    "index": left_fieldname_index_with_suffix,
                },
            ),
            how="inner",
            left_on=left_fieldname_index_with_suffix,
            right_on=left_fieldname_index_with_suffix,
            suffixes=(None, None),
        )
        logger.info("[crossref] Merging intersections with right input file")
        dst_data_frame = dst_data_frame.merge(
            right_data_frame.reset_index().rename(
                index=str,
                columns={
                    "index": right_fieldname_index_with_suffix,
                },
            ),
            how="inner",
            left_on=right_fieldname_index_with_suffix,
            right_on=right_fieldname_index_with_suffix,
            suffixes=(left_suffix, right_suffix),
        )

        # Sort cross-reference results by Jaccard similarity coefficient.
        if sort_by_jaccard:
            logger.info(f"[crossref] Sorting: {fieldname_jaccard!r} {sort_order}")
            dst_data_frame = dst_data_frame.sort_values(by=fieldname_jaccard, axis=0, ascending=sort_order_to_ascending_(sort_order))

        # Group cross-reference results by left "index" and then, for each group,
        # select cross-reference result with least ("ASC") or greatest ("DESC") value.
        if left_group_by_jaccard:
            logger.info(f"[crossref] Grouping on left: {fieldname_jaccard!r} {left_group_order}")
            left_group_by_series = dst_data_frame.groupby([left_fieldname_index_with_suffix])[fieldname_jaccard]
            if left_group_order == "ASC":
                dst_data_frame = dst_data_frame.loc[left_group_by_series.idxmin()]
            elif left_group_order == "DESC":
                dst_data_frame = dst_data_frame.loc[left_group_by_series.idxmax()]
            else:
                pass

        # Group cross-reference results by right "index" and then, for each group,
        # select cross-reference result with least ("ASC") or greatest ("DESC") value.
        if right_group_by_jaccard:
            logger.info(f"[crossref] Grouping on right: {fieldname_jaccard!r} {right_group_order}")
            right_group_by_series = dst_data_frame.groupby([right_fieldname_index_with_suffix])[fieldname_jaccard]
            if right_group_order == "ASC":
                dst_data_frame = dst_data_frame.loc[right_group_by_series.idxmin()]
            elif right_group_order == "DESC":
                dst_data_frame = dst_data_frame.loc[right_group_by_series.idxmax()]
            else:
                pass

        # Delete left and right "__openlocationcode__" fields.
        del dst_data_frame[left_fieldname_openlocationcode_with_suffix]
        del dst_data_frame[right_fieldname_openlocationcode_with_suffix]

        # Delete "IoU" field.
        if not include_jaccard_field:
            del dst_data_frame[fieldname_jaccard]

        # Delete left and right "index" fields.
        if not include_index_fields:
            del dst_data_frame[left_fieldname_index_with_suffix]
            del dst_data_frame[right_fieldname_index_with_suffix]

        # Write output file.
        logger.info(f"[crossref] Writing output file: {dst.name!r}")
        dst_data_frame.to_csv(path_or_buf=dst, header=True, index=False, **kwargs_for_to_csv_dst)
    except BaseException as exception:
        raise click.ClickException(exception)

    # Done!
    return



# Copyright (c) 2019, Battelle Memorial Institute
# All rights reserved.
#
# See LICENSE.txt and WARRANTY.txt for details.

from openlocationcode import openlocationcode


def isValidCodeArea(codeArea: openlocationcode.CodeArea) -> bool:
    return (
        isValidCodeLength(codeArea.codeLength)
        and isValidLatitudeCenter(codeArea.latitudeLo, codeArea.latitudeHi, codeArea.latitudeCenter)
        and isValidLongitudeCenter(codeArea.longitudeLo, codeArea.longitudeHi, codeArea.longitudeCenter)
    )


def isValidCodeLength(codeLength: int) -> bool:
    return (codeLength >= 2) and ((codeLength >= openlocationcode.PAIR_CODE_LENGTH_) or (codeLength % 2 == 0))


def isValidLatitude(latitudeLo: float, latitudeHi: float) -> bool:
    return -openlocationcode.LATITUDE_MAX_ <= latitudeLo <= latitudeHi <= openlocationcode.LATITUDE_MAX_


def isValidLatitudeCenter(latitudeLo: float, latitudeHi: float, latitudeCenter: float) -> bool:
    return -openlocationcode.LATITUDE_MAX_ <= latitudeLo <= latitudeCenter <= latitudeHi <= openlocationcode.LATITUDE_MAX_


def isValidLongitude(longitudeLo: float, longitudeHi: float) -> bool:
    return -openlocationcode.LONGITUDE_MAX_ <= longitudeLo <= longitudeHi <= openlocationcode.LONGITUDE_MAX_


def isValidLongitudeCenter(longitudeLo: float, longitudeHi: float, longitudeCenter: float) -> bool:
    return -openlocationcode.LONGITUDE_MAX_ <= longitudeLo <= longitudeCenter <= longitudeHi <= openlocationcode.LONGITUDE_MAX_



# Copyright (c) 2019, Battelle Memorial Institute
# All rights reserved.
#
# See LICENSE.txt and WARRANTY.txt for details.


import shapely.geometry
import shapely.geometry.point
import shapely.wkb
import shapely.wkt

from buildingid.command_line.dict_datum import DictDatum
from buildingid.command_line.dict_pipe import DictDecoder


class LatLngDictDecoder(DictDecoder[DictDatum]):
    def __init__(
        self,
        fieldname_center_latitude: str,
        fieldname_center_longitude: str,
        fieldname_north_latitude: str | None = None,
        fieldname_south_latitude: str | None = None,
        fieldname_west_longitude: str | None = None,
        fieldname_east_longitude: str | None = None,
    ) -> None:
        super().__init__()

        self.fieldname_center_latitude = fieldname_center_latitude
        self.fieldname_center_longitude = fieldname_center_longitude

        self.fieldname_north_latitude = self.fieldname_center_latitude if (fieldname_north_latitude is None) else fieldname_north_latitude
        self.fieldname_south_latitude = self.fieldname_center_latitude if (fieldname_south_latitude is None) else fieldname_south_latitude
        self.fieldname_east_longitude = self.fieldname_center_longitude if (fieldname_east_longitude is None) else fieldname_east_longitude
        self.fieldname_west_longitude = self.fieldname_center_longitude if (fieldname_west_longitude is None) else fieldname_west_longitude

    def decode(self, row: dict[str, str]) -> DictDatum:
        center_latitude = float(row[self.fieldname_center_latitude])
        center_longitude = float(row[self.fieldname_center_longitude])

        centroid = shapely.geometry.point.Point(center_longitude, center_latitude)

        if (self.fieldname_center_latitude == self.fieldname_north_latitude == self.fieldname_south_latitude) and (
            self.fieldname_center_longitude == self.fieldname_east_longitude == self.fieldname_west_longitude
        ):
            return DictDatum(centroid)
        else:
            north_latitude = float(row[self.fieldname_north_latitude])
            south_latitude = float(row[self.fieldname_south_latitude])
            east_longitude = float(row[self.fieldname_east_longitude])
            west_longitude = float(row[self.fieldname_west_longitude])

            bounds = (west_longitude, south_latitude, east_longitude, north_latitude)

            bbox = shapely.geometry.box(*bounds, ccw=True)

            return DictDatum(bbox, bounds=bounds, centroid=centroid)

    @property
    def fieldnames(self) -> list[str]:
        return list(
            {
                self.fieldname_center_latitude,
                self.fieldname_center_longitude,
                self.fieldname_north_latitude,
                self.fieldname_south_latitude,
                self.fieldname_east_longitude,
                self.fieldname_west_longitude,
            },
        )


class WKBDictDecoder(DictDecoder[DictDatum]):
    def __init__(self, fieldname_wkbstr: str) -> None:
        super().__init__()

        self.fieldname_wkbstr = fieldname_wkbstr

    def decode(self, row: dict[str, str]) -> DictDatum:
        wkbstr = str(row[self.fieldname_wkbstr])

        geom = shapely.wkb.loads(wkbstr, hex=True)

        return DictDatum(geom)

    @property
    def fieldnames(self) -> list[str]:
        return [
            self.fieldname_wkbstr,
        ]


class WKTDictDecoder(DictDecoder[DictDatum]):
    def __init__(self, fieldname_wktstr: str) -> None:
        super().__init__()

        self.fieldname_wktstr = fieldname_wktstr

    def decode(self, row: dict[str, str]) -> DictDatum:
        wktstr = str(row[self.fieldname_wktstr])

        geom = shapely.wkt.loads(wktstr)

        return DictDatum(geom)

    @property
    def fieldnames(self) -> list[str]:
        return [
            self.fieldname_wktstr,
        ]



# Copyright (c) 2019, Battelle Memorial Institute
# All rights reserved.
#
# See LICENSE.txt and WARRANTY.txt for details.

import csv
import sys


# c.f., https://stackoverflow.com/questions/15063936/csv-error-field-larger-than-field-limit-131072
def set_csv_field_size_limit(maxsize: int = sys.maxsize) -> None:
    decrement = True

    while decrement:
        try:
            csv.field_size_limit(maxsize)
        except OverflowError:
            maxsize = int(maxsize / 10)

            decrement = True
        else:
            decrement = False



# Copyright (c) 2019, Battelle Memorial Institute
# All rights reserved.
#
# See LICENSE.txt and WARRANTY.txt for details.


class CustomError(Exception):
    def __init__(self, msg: str) -> None:
        super().__init__()

        self.msg = msg

    def __str__(self) -> str:
        return self.msg


class FieldNotFoundError(CustomError):
    def __init__(self, fieldname: str) -> None:
        msg = f"field {fieldname!r} is not defined"

        super().__init__(msg)

        self.fieldname = fieldname


class FieldNotUniqueError(CustomError):
    def __init__(self, fieldname: str) -> None:
        msg = f"field {fieldname!r} has already been taken"

        super().__init__(msg)

        self.fieldname = fieldname



# Copyright (c) 2019, Battelle Memorial Institute
# All rights reserved.
#
# See LICENSE.txt and WARRANTY.txt for details.

import re

from openlocationcode import openlocationcode

SEPARATOR_ = "-"
FORMAT_STRING_ = "%s-%.0f-%.0f-%.0f-%.0f"

# fmt: off
RE_PATTERN_ = re.compile(
    r"".join(
        [
            r"^",
            r"(",
                r"[", re.escape(openlocationcode.CODE_ALPHABET_[0:9]), r"][", re.escape(openlocationcode.CODE_ALPHABET_[0:18]), r"]",
                r"(?:",
                    re.escape(openlocationcode.PADDING_CHARACTER_), r"{6}",
                    re.escape(openlocationcode.SEPARATOR_),
                    r"(?:",
                        re.escape(openlocationcode.PADDING_CHARACTER_), r"{2,}",
                    r")?",
                    r"|",
                    r"[", re.escape(openlocationcode.CODE_ALPHABET_), r"]{2}",
                    r"(?:",
                        re.escape(openlocationcode.PADDING_CHARACTER_), r"{4}",
                        re.escape(openlocationcode.SEPARATOR_),
                        r"(?:",
                            re.escape(openlocationcode.PADDING_CHARACTER_), r"{2,}",
                        r")?",
                        r"|",
                        r"[", re.escape(openlocationcode.CODE_ALPHABET_), r"]{2}",
                        r"(?:",
                            re.escape(openlocationcode.PADDING_CHARACTER_), r"{2}", re.escape(openlocationcode.SEPARATOR_),
                            r"(?:",
                                re.escape(openlocationcode.PADDING_CHARACTER_), r"{2,}",
                            r")?",
                            r"|",
                            r"[", re.escape(openlocationcode.CODE_ALPHABET_), r"]{2}",
                            re.escape(openlocationcode.SEPARATOR_),
                            r"(?:",
                                re.escape(openlocationcode.PADDING_CHARACTER_), r"{2,}",
                                r"|",
                                r"[", re.escape(openlocationcode.CODE_ALPHABET_), r"]{2,}",
                                re.escape(openlocationcode.PADDING_CHARACTER_), r"*",
                            r")?",
                        r")",
                    r")",
                r")",
            r")",
            re.escape(SEPARATOR_),
            r"(0|[1-9][0-9]*)",
            re.escape(SEPARATOR_),
            r"(0|[1-9][0-9]*)",
            re.escape(SEPARATOR_),
            r"(0|[1-9][0-9]*)",
            re.escape(SEPARATOR_),
            r"(0|[1-9][0-9]*)",
            r"$",
        ],
    ),
    flags=re.IGNORECASE,
)
# fmt: on

RE_GROUP_OPENLOCATIONCODE_ = 1
RE_GROUP_NORTH_ = 2
RE_GROUP_EAST_ = 3
RE_GROUP_SOUTH_ = 4
RE_GROUP_WEST_ = 5
